
# TODO (cg 2022/07/12 12:05): Break out a different function that prepares the
# necessary functions, and then write another version of ipopt_optimize that
# accepts all the functions? I don't see much of a point in that because that
# doesn't actually have all that much boilerplate anymore. On the other hand,
# there are a lot of ifs all over this now. I feel like breaking this up to
# handle sparsity and mixed-mode AD would make this code much more readable.
function ipopt_optimize(obj, ini, constraints=noconstraints(); 
                        jac_sparsity::SPBOOL_OR_NOTHING=nothing, 
                        mixed_ad=false, return_internals=false, 
                        box_lower=-1e22, box_upper=1e22, kwargs...)
  densej  = isnothing(jac_sparsity)
  # bring in some default arguments:
  optargs = merge(DEF_IPOPT_KWARGS, Dict(kwargs))
  # get the dimension of the problem:
  arglen  = length(ini)
  # color vector for the jacobian if it is sparse:
  jcolorv = densej ? nothing : matrix_colors(jac_sparsity) 
  # set up box constraints for the args and, if applicable, the constraints:
  box_l = prepare_bounds(arglen, box_lower)
  box_u = prepare_bounds(arglen, box_upper)
  # function for the constraints:
  constr_buf  = zeros(constraints.ncon)
  # If mixed AD is requested, create that object:
  if mixed_ad
    obj_wrapper = FoRWrapper(obj, ini)
  end
  # objective gradient and constraint jacobian:
  if mixed_ad
    objg = (p,g)->begin
      grad = obj_wrapper(:gradient, p)
      g .= grad
    end
  else
    objg = (p,g)->ForwardDiff.gradient!(g, obj, p)
  end
  jac_buf = densej ? zeros(constraints.ncon, arglen) : Float64.(jac_sparsity)
  jac_nz  = densej ? arglen*constraints.ncon : sum(!iszero, jac_sparsity)
  if densej 
    constr_jac = (p,r,c,v) -> begin 
      isnothing(v) && return jac_structure!(r, c, constraints.ncon, length(p)) 
      ForwardDiff.jacobian!(jac_buf, constraints.fun, constr_buf, p) 
      v .= vec(jac_buf') 
    end
  else
    constr_jac = (p,r,c,v) -> begin 
      isnothing(v) && return jac_structure_sparse!(jac_sparsity, r, c)
      forwarddiff_color_jacobian!(jac_buf, constraints.fun, p, colorvec=jcolorv)
      populate_sparse_jacobian!(v, jac_sparsity, jac_buf)
    end
  end
  # Lagrangian hessian:
  if constraints.ncon > 0
    lagrange = (p,l,o) -> begin
      conbuf = zeros(eltype(p), length(constraints.box_lower))
      constraints.fun(conbuf, p)
      obj(p)*o + dot(l, conbuf)
    end
  else
    lagrange = (p,l,o) -> obj(p)*o 
  end
  lagrange_h = (p,r,c,o,l,v) -> begin 
    isnothing(v) && return hes_structure!(r,c,length(p)) 
    if mixed_ad
      lagrange_wrapper = FoRWrapper(_p->lagrange(_p, l, o), p)
      lhess = lagrange_wrapper(:hessian, p)
    else
      lhess = ForwardDiff.hessian(_p->lagrange(_p, l, o), p)
    end
    linear_buffer_ltri_colmajor!(v, lhess) 
    nothing
  end
  # For debugging:
  return_internals && return (;obj, constraints, objg, constr_jac, lagrange_h)
  # Create the final problem:
  constr_eval = (p, store) -> constraints.fun(store, p)
  prob = CreateIpoptProblem(arglen,  box_l, box_u,     # arg count and box
                            constraints.ncon, 
                            constraints.box_lower, 
                            constraints.box_upper,     # constr. count and box
                            jac_nz,                    # size of constr. jacobian
                            div(arglen*(arglen+1), 2), # size of hessian ltri
                            obj,                       # objective
                            constr_eval,               # constraint function
                            objg,                      # objective gradient
                            constr_jac,                # constraint jacobian
                            lagrange_h)                # lagrangian hessian
  # Add all the options:
  for (opt, val) in optargs
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=status, minimizer=copy(prob.x), minval=obj(prob.x))
end

function ipopt_nlsolve(fn, ini; box_lower=-1e22, box_upper=1e22, kwargs...)
  # bring in some default arguments:
  optargs = merge(DEF_IPOPT_KWARGS, Dict(kwargs))
  # get the dimension of the problem:
  arglen  = length(ini)
  nconstr = arglen
  # set up box constraints for the args and, if applicable, the constraints:
  box_l = prepare_bounds(arglen, box_lower)
  box_u = prepare_bounds(arglen, box_upper)
  con_l = zeros(length(ini))
  con_u = zeros(length(ini))
  # function for the constraints:
  constr_eval = (p,store) -> fn(store, p)
  # "constraint" jacobian:
  con_buf = zeros(arglen)
  jac_buf = zeros(arglen, arglen)
  constr_jac = (p,r,c,v) -> begin 
    isnothing(v) && return jac_structure!(r, c, nconstr, nconstr) 
    ForwardDiff.jacobian!(jac_buf, fn, con_buf, p)
    v .= vec(jac_buf)
  end
  # Lagrangian hessian:
  lagrange_h = (p,r,c,o,l,v) -> begin 
    isnothing(v) && return hes_structure!(r,c,length(p)) 
    fill!(v, 0.0)
    nothing
  end
  # Create the final problem:
  prob = CreateIpoptProblem(arglen,  box_l, box_u,        # arg count and box
                            nconstr, con_l, con_u,        # constr. count and box
                            arglen^2,                     # size of constr. jacobian
                            div(arglen*(arglen+1), 2),    # size of hessian ltri
                            _p->0.0,                      # objective
                            constr_eval,                  # constraint function
                            (p,g)->(g.=zeros(length(p))), # objective gradient
                            constr_jac,                   # constraint jacobian
                            lagrange_h)                   # lagrangian hessian
  # Add all the options:
  ipopt_addoption!(prob, "hessian_approximation", "limited-memory")
  ipopt_addoption!(prob, "nlp_scaling_method",    "none")
  for (opt, val) in optargs
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=status, minimizer=copy(prob.x))
end

